r"""Lattice sums.

.. currentmodule:: treams.lattice

Calculates the lattice sums of the forms

.. math::
   D_{lm}(k, \boldsymbol k_\parallel, \boldsymbol r, \Lambda_d)
   = \sideset{}{'}{\sum_{\boldsymbol R \in \Lambda_d}}
   h_l^{(1)}(k|\boldsymbol r + \boldsymbol R|)
   Y_{lm}
   (\theta_{-\boldsymbol r - \boldsymbol R}, \varphi_{-\boldsymbol r - \boldsymbol R})
   \mathrm e^{\mathrm i \boldsymbol k_\parallel \boldsymbol R}

and

.. math::
   D_{m}(k, \boldsymbol k_\parallel, \boldsymbol r, \Lambda_d)
   = \sideset{}{'}{\sum_{\boldsymbol R \in \Lambda_d}}
   H_m^{(1)}(k|\boldsymbol r + \boldsymbol R|)
   \mathrm e^{\mathrm i m \varphi_{-\boldsymbol r - \boldsymbol R}}
   \mathrm e^{\mathrm i \boldsymbol k_\parallel \boldsymbol R}

that arise when translating and summing the spherical and cylindrical solutions
of the Helmholtz equation in a periodic arrangement. These sums have a notoriously slow
convergence and at least for lattices with a dimension :math:`d > 1` it is not advisable
to use the direct approach. Fortunately, it is possible to convert them to exponentially
convergent series, which are implemented here. For details on the math, see the
references below.

The lattice of dimension :math:`d \leq 3` (:math:`d \leq 2` in the second case) is
denoted with :math:`\Lambda_d` and consists of all vectors
:math:`\boldsymbol R = \sum_{i=1}^{d} n_i \boldsymbol a_i` with :math:`\boldsymbol a_i`
being basis vectors of the lattice and :math:`n_i \in \mathbb Z`. For :math:`d = 2` the
lattice is in the `z = 0` plane and for :math:`d = 1` it is along the z-axis. The vector
:math:`\boldsymbol r` is arbitrary but for best convergence it should be reduced to the
Wigner-Seitz cell of the lattice. The summation excludes the point
:math:`\boldsymbol R + \boldsymbol r = 0` if it exists, which is indicated by the prime
next to the summation sign.

The wave in the lattice is defined by its -- possibly complex-valued -- wave number
:math:`k` and the (real) components of the wave vector
:math:`\boldsymbol k_\parallel \in \mathbb R^d` that are parallel to the lattice.

The expressions include the (spherical) Hankel functions of first kind :math:`H_m^{(1)}`
(:math:`h_l^{(1)}`) and the spherical harmonics :math:`Y_{lm}`. The angles
:math:`\theta` and :math:`\varphi` are the polar and azimuthal angle, when expressing
the points in spherical coordinates. In the first case, the degree is
:math:`l \in \mathbb N_0` and the order is :math:`\mathbb Z \ni m \leq l`. In the second
case :math:`m \in \mathbb Z`.


Available functions
===================

Accelerated lattice summations
------------------------------

.. autosummary::
   :toctree:

   lsumcw1d
   lsumcw1d_shift
   lsumcw2d
   lsumsw1d
   lsumsw1d_shift
   lsumsw2d
   lsumsw2d_shift
   lsumsw3d


Direct summations
-----------------

The functions are almost only here for benchmarking and comparison.

.. autosummary::
   :toctree:

   dsumcw1d
   dsumcw1d_shift
   dsumcw2d
   dsumsw1d
   dsumsw1d_shift
   dsumsw2d
   dsumsw2d_shift
   dsumsw3d


Miscellaneous functions
-----------------------

.. autosummary::
   :toctree:

   area
   cube
   cubeedge
   diffr_orders_circle
   reciprocal
   volume


Cython module
-------------

.. autosummary::
   :toctree:
   :template: cython-module

   cython_lattice


References
==========

* `[1] K. Kambe, Zeitschrift Für Naturforschung A 22, 3 (1967). <https://doi.org/10.1515/zna-1967-0305>`_
* `[2] K. Kambe, Zeitschrift Für Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
* `[3] C. M. Linton, SIAM Rev. 52, 630 (2010). <https://doi.org/10.1137/09075130X>`_
* `[4] D. Beutel el al., J. Opt. Soc. Am. B (2021). <https://doi.org/10.1364/JOSAB.419645>`_
"""

import numpy as np

from treams.lattice import _misc
from treams.lattice._gufuncs import *  # noqa: F403
from treams.lattice._misc import cube, cubeedge  # noqa: F401


def diffr_orders_circle(b, rmax):
    """Diffraction orders in a circle.

    Given the reciprocal lattice defined by the vectors that make up the rows of `b`,
    return all diffraction orders within a circle of radius `rmax`.

    Args:
        b (float, (2, 2)-array): Reciprocal lattice vectors
        rmax (float): Maximal radius

    Returns:
        float array
    """
    return _misc.diffr_orders_circle(np.array(b), rmax)


def lsumsw(dim, l, m, k, kpar, a, r, eta, out=None, **kwargs):  # noqa: E741
    if dim == 1:
        return lsumsw1d_shift(l, m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return lsumsw2d_shift(l, m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    if dim == 3:
        return lsumsw3d(l, m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def realsumsw(dim, l, m, k, kpar, a, r, eta, out=None, **kwargs):  # noqa: E741
    if dim == 1:
        return realsumsw1d_shift(  # noqa: F405
            l, m, k, kpar, a, r, eta, out=out, **kwargs
        )
    if dim == 2:
        return realsumsw2d_shift(  # noqa: F405
            l, m, k, kpar, a, r, eta, out=out, **kwargs
        )
    if dim == 3:
        return realsumsw3d(l, m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def recsumsw(dim, l, m, k, kpar, a, r, eta, out=None, **kwargs):  # noqa: E741
    if dim == 1:
        return recsumsw1d_shift(  # noqa: F405
            l, m, k, kpar, a, r, eta, out=out, **kwargs
        )
    if dim == 2:
        return recsumsw2d_shift(  # noqa: F405
            l, m, k, kpar, a, r, eta, out=out, **kwargs
        )
    if dim == 3:
        return recsumsw3d(l, m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def lsumcw(dim, m, k, kpar, a, r, eta, out=None, **kwargs):  # noqa: E741
    if dim == 1:
        return lsumcw1d_shift(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return lsumcw2d(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def realsumcw(dim, m, k, kpar, a, r, eta, out=None, **kwargs):
    if dim == 1:
        return realsumcw1d_shift(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return realsumcw2d(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def recsumcw(dim, m, k, kpar, a, r, eta, out=None, **kwargs):
    if dim == 1:
        return recsumcw1d_shift(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return recsumcw2d(m, k, kpar, a, r, eta, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def dsumsw(dim, l, m, k, kpar, a, r, i, out=None, **kwargs):  # noqa: E741
    if dim == 1:
        return dsumsw1d_shift(l, m, k, kpar, a, r, i, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return dsumsw2d_shift(l, m, k, kpar, a, r, i, out=out, **kwargs)  # noqa: F405
    if dim == 3:
        return dsumsw3d(l, m, k, kpar, a, r, i, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")


def dsumcw(dim, m, k, kpar, a, r, i, out=None, **kwargs):
    if dim == 1:
        return dsumcw1d_shift(m, k, kpar, a, r, i, out=out, **kwargs)  # noqa: F405
    if dim == 2:
        return dsumcw2d(m, k, kpar, a, r, i, out=out, **kwargs)  # noqa: F405
    raise ValueError(f"invalid dimension '{dim}'")



ctypedef fused number_t:
    double complex
    double


cpdef double complex hankel1_d(double l, double complex z) nogil
cpdef double complex hankel2_d(double l, double complex z) nogil
cpdef number_t jv_d(double l, number_t z) nogil
cpdef double complex spherical_hankel1(double n, number_t z) nogil
cpdef double complex spherical_hankel1_d(double l, double complex z) nogil
cpdef double complex spherical_hankel2(double n, number_t z) nogil
cpdef double complex spherical_hankel2_d(double l, double complex z) nogil
cpdef number_t yv_d(double l, number_t z) nogil


cdef void car2cyl(double *input, double *output) nogil
cdef void car2pol(double *input, double *output) nogil
cdef void car2sph(double *input, double *output) nogil
cdef void cyl2car(double *input, double *output) nogil
cdef void cyl2sph(double *input, double *output) nogil
cdef void pol2car(double *input, double *output) nogil
cdef void sph2car(double *input, double *output) nogil
cdef void sph2cyl(double *input, double *output) nogil

cdef void vcar2cyl(number_t *iv, double *ip, number_t *ov) nogil
cdef void vcar2pol(number_t *iv, double *ip, number_t *ov) nogil
cdef void vcar2sph(number_t *iv, double *ip, number_t *ov) nogil
cdef void vcyl2car(number_t *iv, double *ip, number_t *ov) nogil
cdef void vcyl2sph(number_t *iv, double *ip, number_t *ov) nogil
cdef void vpol2car(number_t *iv, double *ip, number_t *ov) nogil
cdef void vsph2car(number_t *iv, double *ip, number_t *ov) nogil
cdef void vsph2cyl(number_t *iv, double *ip, number_t *ov) nogil


cpdef number_t incgamma(double l, number_t z) nogil
cpdef number_t intkambe(long n, number_t z, number_t eta) nogil


cpdef number_t lpmv(double m, double l, number_t z) nogil
cpdef number_t pi_fun(double l, double m, number_t x) nogil
cpdef double complex sph_harm(double m, double l, double phi, number_t theta) nogil
cpdef number_t tau_fun(double l, double m, number_t x) nogil

cpdef double complex tl_vcw(double kz1, long mu, double kz2, long m, double complex krr, double phi, double z) nogil
cpdef double complex tl_vcw_r(double kz1, long mu, double kz2, long m, number_t krr, double phi, double z) nogil

cpdef double complex _tl_vsw_helper(long l, long m, long lambda_, long mu, long p, long q) nogil

cpdef double complex tl_vsw_A(long lambda_, long mu, long l, long m, double complex kr, number_t theta, double phi) nogil
cpdef double complex tl_vsw_B(long lambda_, long mu, long l, long m, double complex kr, number_t theta, double phi) nogil
cpdef double complex tl_vsw_rA(long lambda_, long mu, long l, long m, number_t kr, number_t theta, double phi) nogil
cpdef double complex tl_vsw_rB(long lambda_, long mu, long l, long m, number_t kr, number_t theta, double phi) nogil

cdef void vcw_A(double kz, long m, double complex krr, double phi, double z, double complex k, long pol, double complex *out) nogil
cdef void vcw_M(double kz, long m, double complex krr, double phi, double z, double complex *out) nogil
cdef void vcw_N(double kz, long m, double complex krr, double phi, double z, double complex k, double complex *out) nogil
cdef void vcw_rA(double kz, long m, number_t krr, double phi, double z, double complex k, long pol, double complex *out) nogil
cdef void vcw_rM(double kz, long m, number_t krr, double phi, double z, double complex *out, long i) nogil
cdef void vcw_rN(double kz, long m, number_t krr, double phi, double z, double complex k, double complex *out) nogil

cdef void vpw_A(number_t kx, number_t ky, number_t kz, double x, double y, double z, long pol, double complex *out) nogil
cdef void vpw_M(number_t kx, number_t ky, number_t kz, double x, double y, double z, double complex *out) nogil
cdef void vpw_N(number_t kx, number_t ky, number_t kz, double x, double y, double z, double complex *out) nogil

cdef void vsh_X(long l, long m, number_t theta, double phi, double complex *out) nogil
cdef void vsh_Y(long l, long m, number_t theta, double phi, double complex *out) nogil
cdef void vsh_Z(long l, long m, number_t theta, double phi, double complex *out) nogil

cdef void vsw_A(long l, long m, double complex kr, number_t theta, double phi, long pol, double complex *out) nogil
cdef void vsw_M(long l, long m, double complex kr, number_t theta, double phi, double complex *out) nogil
cdef void vsw_N(long l, long m, double complex kr, number_t theta, double phi, double complex *out) nogil
cdef void vsw_rA(long l, long m, number_t kr, number_t theta, double phi, long pol, double complex *out) nogil
cdef void vsw_rM(long l, long m, number_t kr, number_t theta, double phi, double complex *out) nogil
cdef void vsw_rN(long l, long m, number_t kr, number_t theta, double phi, double complex *out) nogil


cpdef number_t wignersmalld(long l, long m, long k, number_t theta) nogil
cpdef double complex wignerd(long l, long m, long k, double phi, number_t theta, double psi) nogil


cpdef double wigner3j(long j1, long j2, long j3, long m1, long m2, long m3) nogil



"""Operators for common transformations including different types of waves."""

import inspect

import numpy as np

import treams._core as core
import treams.special as sc
from treams import config, cw, pw, sw, util
from treams._lattice import Lattice, WaveVector
from treams._material import Material


class Operator:
    """Operator base class.

    An operator is mainly intended to be used as a descriptor on a class. It will then
    automatically obtain attributes from the class that correspond to the variable names
    of the function that is linked to the operator in the class attribute `_FUNC`. When
    called, the remaining arguments have to be specified.
    """

    __array_priority__ = 1.0
    _FUNC = None

    def __init__(self, *args, isinv=False):
        self._args = args
        self._isinv = bool(isinv)

    @property
    def isinv(self):
        return self._isinv

    @property
    def inv(self):
        return type(self)(*self._args, isinv=not self.isinv)

    @property
    def FUNC(self):
        return self._FUNC

    def __call__(self, **kwargs):
        if self.isinv:
            return self._call_inv(**kwargs)
        return self.FUNC(*self._args, **kwargs)

    def _call_inv(self, **kwargs):
        return self.FUNC(*self._args, **kwargs)

    def __matmul__(self, other):
        if isinstance(other, Operator):
            raise NotImplementedError
        return self(**self.get_kwargs(other, max(-2, -np.ndim(other)))) @ other

    def __rmatmul__(self, other):
        return other @ self(**self.get_kwargs(other))

    def get_kwargs(self, obj, dim=-1):
        kwargs = {}
        for name, param in inspect.signature(self.FUNC).parameters.items():
            if param.kind == param.KEYWORD_ONLY:
                val = getattr(obj, name, None)
                if isinstance(val, tuple):
                    val = val[dim]
                if val is not None:
                    kwargs[name] = val
        return kwargs

    def __repr__(self):
        if len(self._args) == 1:
            args = f"({self._args[0]})"
        else:
            args = str(self._args)
        if self.isinv:
            args = args[:-1] + ", isinv=True)"
        return f"{type(self).__name__}" + args


class OperatorAttribute:
    def __init__(self, op):
        self._op = op
        self._obj = self._objtype = None

    @property
    def OP(self):
        return self._op

    def eval(self, *args, **kwargs):
        op = self.OP(*args)
        dim = max(-2, -np.ndim(self._obj))
        return op(**self._merge_kwargs(kwargs, op.get_kwargs(self._obj, dim)))

    @staticmethod
    def _merge_kwargs(kwargsa, kwargsb):
        for key, val in kwargsb.items():
            if key == "basis" and key in kwargsa:
                kwargsa[key] = kwargsa["basis"], val
            elif key in kwargsa:
                raise TypeError(f"got multiple values for keyword argument '{key}'")
            else:
                kwargsa[key] = val
        return kwargsa

    def eval_inv(self, *args, **kwargs):
        op = self.OP(*args).inv
        return op(**self._merge_kwargs(kwargs, op.get_kwargs(self._obj)))

    def __get__(self, obj, objtype=None):
        self._obj = obj
        self._objtype = type(obj) if objtype is None else objtype
        return self

    def __call__(self, *args, **kwargs):
        if np.ndim(self._obj) == 1:
            return self.apply_left(*args, **kwargs)
        try:
            inv = self.eval_inv(*args, **kwargs)
        except NotImplementedError:
            return self.apply_left(*args, **kwargs)
        if issubclass(self._objtype, util.AnnotatedArray):
            return self._objtype.relax(self.eval(*args, **kwargs) @ self._obj @ inv)
        return self.eval(*args, **kwargs) @ self._obj @ inv

    def apply_left(self, *args, **kwargs):
        return self.eval(*args, **kwargs) @ self._obj

    def apply_right(self, *args, **kwargs):
        return self._obj @ self.eval_inv(*args, **kwargs)

    def __repr__(self):
        return f"{type(self).__name__}({self.OP.__name__})"


def _sw_rotate(phi, theta, psi, basis, to_basis, where):
    """Rotate spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    res = sw.rotate(
        *(m[:, None] for m in to_basis.lms), *basis.lms, phi, theta, psi, where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(res, basis=(to_basis, basis))


def _cw_rotate(phi, basis, to_basis, where):
    """Rotate cylindrical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    res = cw.rotate(*(m[:, None] for m in to_basis.zms), *basis.zms, phi, where=where,)
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(res, basis=(to_basis, basis))


def _pwa_rotate(phi, basis, where):
    """Rotate plane waves (actually rotates the basis)."""
    c1, s1 = np.cos(phi), np.sin(phi)
    r = np.array([[c1, -s1, 0], [s1, c1, 0], [0, 0, 1]])
    kvecs = r @ np.array([basis.qx, basis.qy, basis.qz])
    res = np.eye(len(basis))
    res[..., np.logical_not(where)] = 0
    newbasis = core.PlaneWaveBasisByUnitVector(zip(*kvecs, basis.pol))
    if basis.lattice is not None:
        newbasis.lattice = basis.lattice.rotate(phi)
    if basis.kpar is not None:
        newbasis.kpar = basis.kpar.rotate(phi)
    return core.PhysicsArray(res, basis=(newbasis, basis))


def _pwp_rotate(phi, basis, where):
    """Rotate partial plane waves (actually rotates the basis)."""
    if basis.alignment != "xy":
        raise ValueError(f"rotation on alignment: '{basis.alignment}'")
    c1, s1 = np.cos(phi), np.sin(phi)
    r = np.array([[c1, -s1], [s1, c1]])
    kx, ky, pol = basis[()]
    res = np.eye(len(basis))
    res[..., np.logical_not(where)] = 0
    newbasis = core.PlaneWaveBasisByComp(zip(*(r @ np.array([kx, ky])), pol))
    if basis.lattice is not None:
        newbasis.lattice = basis.lattice.rotate(phi)
    if basis.kpar is not None:
        newbasis.kpar = basis.kpar.rotate(phi)
    return core.PhysicsArray(res, basis=(newbasis, basis))


def rotate(phi, theta=0, psi=0, *, basis, where=True):
    """Rotation matrix.

    For the given Euler angles apply a rotation for the given basis. In some basis sets
    only rotations around the z-axis are permitted.

    Args:
        phi (float): First Euler angle (rotation about z)
        theta (float, optional): Second Euler angle (rotation about y)
        psi (float, optional): Third Euler angle (rotation about z)
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        where (array-like, bool, optional): Only evaluate parts of the rotation matrix,
            the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis

    if isinstance(basis, core.SphericalWaveBasis):
        return _sw_rotate(phi, theta, psi, basis, to_basis, where)
    if theta != 0:
        raise ValueError("non-zero theta for rotation")
    phi = phi + psi
    if isinstance(basis, core.CylindricalWaveBasis):
        return _cw_rotate(phi, basis, to_basis, where)
    if to_basis != basis:
        raise ValueError("invalid basis")
    if isinstance(basis, core.PlaneWaveBasisByComp):
        return _pwp_rotate(phi, basis, where)
    if isinstance(basis, core.PlaneWaveBasisByUnitVector):
        return _pwa_rotate(phi, basis, where)
    raise TypeError("invalid basis")


class Rotate(Operator):
    _FUNC = staticmethod(rotate)

    def __init__(self, phi, theta=0, psi=0, *, isinv=False):
        super().__init__(phi, theta, psi, isinv=isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs and isinstance(kwargs["basis"], tuple):
            kwargs["basis"] = kwargs["basis"][::-1]
        return self.FUNC(*(-a for a in self._args[::-1]), **kwargs)


def _sw_translate(r, basis, to_basis, k0, material, poltype, where):
    """Translate spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = k0 * material.nmp
    r = sc.car2sph(r)
    res = sw.translate(
        *(m[:, None] for m in to_basis.lms),
        *basis.lms,
        ks[basis.pol] * r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        singular=False,
        poltype=poltype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=(k0, k0),
        basis=(to_basis, basis),
        poltype=(poltype, poltype),
        material=(material, material),
    )


def _cw_translate(r, basis, k0, to_basis, material, where):
    """Translate cylindrical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = material.ks(k0)[basis.pol]
    krhos = np.sqrt(ks * ks - basis.kz * basis.kz + 0j)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    r = sc.car2cyl(r)
    res = cw.translate(
        *(m[:, None] for m in to_basis.zms),
        *basis.zms,
        krhos * r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        singular=False,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res, k0=(k0, k0), basis=(to_basis, basis), material=(material, material),
    )


def _pw_translate(r, basis, k0, to_basis, material, modetype, where):
    """Translate plane waves."""
    kvecs = basis.kvecs(k0, material, modetype)
    kx, ky, kz = to_basis.kvecs(k0, material, modetype)
    where = (
        where
        & (np.abs(kx[:, None] - kvecs[0]) < 1e-14)
        & (np.abs(ky[:, None] - kvecs[1]) < 1e-14)
        & (np.abs(kz[:, None] - kvecs[2]) < 1e-14)
        & (to_basis.pol[:, None] == basis.pol)
    )
    res = pw.translate(
        *kvecs,
        r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=(k0,) * 2,
        basis=(to_basis, basis),
        material=(material,) * 2,
        modetype=(modetype,) * 2,
    )


def translate(
    r, *, basis, k0=None, material=Material(), modetype=None, poltype=None, where=True
):
    """Translation matrix.

    Translate the given basis modes along the translation vector.

    Args:
        r (array-like): Translation vector
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode, only used for
            :class:`~treams.PlaneWaveBasisByComp`.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
        where (array-like, bool, optional): Only evaluate parts of the translation
            matrix, the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    poltype = config.POLTYPE if poltype is None else poltype
    material = Material(material)

    r = np.asanyarray(r)
    if r.shape[-1] != 3:
        raise ValueError("invalid 'r'")

    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_translate(r, basis, k0, to_basis, material, modetype, where)
    if isinstance(basis, core.SphericalWaveBasis):
        return _sw_translate(r, basis, to_basis, k0, material, poltype, where)
    if isinstance(basis, core.CylindricalWaveBasis):
        return _cw_translate(r, basis, k0, to_basis, material, where)
    raise TypeError("invalid basis")


class Translate(Operator):
    """Translation matrix.

    When called as attribute of an object it returns a suitable translation matrix to
    transform it. See also :func:`translate`.
    """

    _FUNC = staticmethod(translate)

    def __init__(self, r, *, isinv=False):
        super().__init__(r, isinv=isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs and isinstance(kwargs["basis"], tuple):
            kwargs["basis"] = kwargs["basis"][::-1]
        return self.FUNC(np.negative(self._args[0]), **kwargs)


def _sw_changepoltype(basis, to_basis, poltype, where):
    """Change the polarization type of spherical waves."""
    where = (
        (to_basis.l[:, None] == basis.l)
        & (to_basis.m[:, None] == basis.m)
        & (to_basis.pidx[:, None] == basis.pidx)
        & where
    )
    res = np.zeros_like(where, float)
    res[where] = np.sqrt(0.5)
    res[where & (to_basis.pol[:, None] == basis.pol) & (basis.pol == 0)] = -np.sqrt(0.5)
    return core.PhysicsArray(res, basis=(to_basis, basis), poltype=poltype)


def _cw_changepoltype(basis, to_basis, poltype, where):
    """Change the polarization type of cylindrical waves."""
    where = (
        (to_basis.kz[:, None] == basis.kz)
        & (to_basis.m[:, None] == basis.m)
        & (to_basis.pidx[:, None] == basis.pidx)
        & where
    )
    res = np.zeros_like(where, float)
    res[where] = np.sqrt(0.5)
    res[where & (to_basis.pol[:, None] == basis.pol) & (basis.pol == 0)] = -np.sqrt(0.5)
    return core.PhysicsArray(res, basis=(to_basis, basis), poltype=poltype)


def _pwa_changepoltype(basis, to_basis, poltype, where):
    """Change the polarization type of plane waves."""
    where = (
        (to_basis.qx[:, None] == basis.qx)
        & (to_basis.qy[:, None] == basis.qy)
        & (to_basis.qz[:, None] == basis.qz)
        & where
    )
    res = np.zeros_like(where, float)
    res[where] = np.sqrt(0.5)
    res[where & (to_basis.pol[:, None] == basis.pol) & (basis.pol == 0)] = -np.sqrt(0.5)
    return core.PhysicsArray(res, basis=(to_basis, basis), poltype=poltype)


def _pwp_changepoltype(basis, to_basis, poltype, where):
    """Change the polarization type of partial plane waves."""
    if to_basis.alignment != basis.alignment:
        raise ValueError("incompatible basis alignments")
    akx, aky, _ = basis[()]
    bkx, bky, _ = to_basis[()]
    where = (bkx[:, None] == akx) & (bky[:, None] == aky) & where
    res = np.zeros_like(where, float)
    res[where] = np.sqrt(0.5)
    res[where & (to_basis.pol[:, None] == basis.pol) & (basis.pol == 0)] = -np.sqrt(0.5)
    return core.PhysicsArray(res, basis=(to_basis, basis), poltype=poltype)


def changepoltype(poltype=None, *, basis, where=True):
    """Matrix to change polarization types.

    The polarization is switched between `helicity` and `parity`.

    Args:
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        where (array-like, bool, optional): Only evaluate parts of the matrix, the given
            array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    poltype = config.POLTYPE if poltype is None else poltype
    if poltype == "helicity":
        poltype = ("helicity", "parity")
    elif poltype == "parity":
        poltype = ("parity", "helicity")
    if poltype not in (("helicity", "parity"), ("parity", "helicity")):
        raise ValueError(f"invalid poltype '{poltype}'")
    if isinstance(basis, core.SphericalWaveBasis):
        return _sw_changepoltype(basis, to_basis, poltype, where)
    if isinstance(basis, core.CylindricalWaveBasis):
        return _cw_changepoltype(basis, to_basis, poltype, where)
    if isinstance(basis, core.PlaneWaveBasisByComp):
        return _pwp_changepoltype(basis, to_basis, poltype, where)
    if isinstance(basis, core.PlaneWaveBasisByUnitVector):
        return _pwa_changepoltype(basis, to_basis, poltype, where)
    raise TypeError("invalid basis")


class ChangePoltype(Operator):
    """Matrix to change polarization types.

    When called as attribute of an object it returns a suitable matrix to change the
    polarization types between `helicity` and `parity`. See also :func:`changepoltype`.
    """

    _FUNC = staticmethod(changepoltype)

    def __init__(self, poltype=None, *, isinv=False):
        args = () if poltype is None else (poltype,)
        super().__init__(*args, isinv=isinv)

    def get_kwargs(self, obj, dim=-1):
        kwargs = super().get_kwargs(obj, dim)
        if self._args:
            return kwargs
        val = getattr(obj, "poltype", None)
        if isinstance(val, tuple):
            val = val[dim]
        opp = {"parity": "helicity", "helicity": "parity"}
        if val is not None:
            kwargs["poltype"] = opp[val]
        return kwargs

    def _call_inv(self, **kwargs):
        if "basis" in kwargs and isinstance(kwargs["basis"], tuple):
            kwargs["basis"] = kwargs["basis"][::-1]
        opp = {"parity": "helicity", "helicity": "parity"}
        poltype = self._args[0] if self._args else kwargs.pop("poltype")
        if isinstance(poltype, tuple):
            poltype = poltype[::-1]
        else:
            poltype = opp[poltype]
        return self.FUNC(poltype, **kwargs)


def _sw_sw_expand(basis, to_basis, to_modetype, k0, material, modetype, poltype, where):
    """Expand spherical waves in spherical waves."""
    if not (
        modetype == "regular" == to_modetype
        or modetype == "singular" == to_modetype
        or (modetype == "singular" and to_modetype == "regular")
    ):
        raise ValueError(f"invalid expansion from {modetype} to {to_modetype}")
    rs = sc.car2sph(to_basis.positions[:, None, :] - basis.positions)
    ks = k0 * material.nmp
    res = sw.translate(
        *(m[:, None] for m in to_basis.lms),
        *basis.lms,
        ks[basis.pol] * rs[to_basis.pidx[:, None], basis.pidx, 0],
        rs[to_basis.pidx[:, None], basis.pidx, 1],
        rs[to_basis.pidx[:, None], basis.pidx, 2],
        poltype=poltype,
        singular=modetype != to_modetype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    res = core.PhysicsArray(
        res, k0=k0, basis=(to_basis, basis), poltype=poltype, material=material
    )
    if modetype == "singular" and to_modetype == "regular":
        res.modetype = (to_modetype, modetype)
    return res


def _sw_cw_expand(basis, to_basis, k0, material, poltype, where):
    """Expand cylindrical waves in spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = material.ks(k0)[basis.pol]
    res = cw.to_sw(
        *(m[:, None] for m in to_basis.lms),
        *basis.zms,
        ks,
        poltype=poltype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        poltype=poltype,
        material=material,
        modetype=("regular", "regular"),
    )


def _sw_pw_expand(basis, to_basis, k0, material, modetype, poltype, where):
    """Expand plane waves in spherical waves."""
    if isinstance(basis, core.PlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    res = pw.to_sw(
        *(m[:, None] for m in to_basis.lms),
        *kvecs,
        basis.pol,
        poltype=poltype,
        where=where,
    ) * pw.translate(
        *kvecs,
        to_basis.positions[to_basis.pidx, None, 0],
        to_basis.positions[to_basis.pidx, None, 1],
        to_basis.positions[to_basis.pidx, None, 2],
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        material=material,
        poltype=poltype,
        modetype=("regular", modetype),
    )


def _cw_cw_expand(basis, to_basis, to_modetype, k0, material, modetype, where):
    """Expand cylindrical waves in cylindrical waves."""
    if modetype == "regular" == to_modetype or modetype == "singular" == to_modetype:
        modetype = to_modetype = None
    elif modetype != "singular" or to_modetype != "regular":
        raise ValueError(f"invalid expansion from {modetype} to {to_modetype}")
    rs = sc.car2cyl(to_basis.positions[:, None, :] - basis.positions)
    krhos = material.krhos(k0, basis.kz, basis.pol)
    res = cw.translate(
        *(m[:, None] for m in to_basis.zms),
        *basis.zms,
        krhos * rs[to_basis.pidx[:, None], basis.pidx, 0],
        rs[to_basis.pidx[:, None], basis.pidx, 1],
        rs[to_basis.pidx[:, None], basis.pidx, 2],
        singular=modetype != to_modetype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    res = core.PhysicsArray(res, k0=k0, basis=(to_basis, basis), material=material)
    if modetype == "singular" and to_modetype == "regular":
        res.modetype = (to_modetype, modetype)
    return res


def _cw_pw_expand(basis, to_basis, k0, material, modetype, where):
    """Expand plane waves in cylindrical waves."""
    if isinstance(basis, core.PlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    res = pw.to_cw(
        *(m[:, None] for m in to_basis.zms), *kvecs, basis.pol, where=where,
    ) * pw.translate(
        *kvecs,
        to_basis.positions[to_basis.pidx, None, 0],
        to_basis.positions[to_basis.pidx, None, 1],
        to_basis.positions[to_basis.pidx, None, 2],
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        material=material,
        modetype=("regular", modetype),
    )


def _pw_pw_expand(basis, to_basis, k0, material, modetype, where):
    """Expand plane waves in plane waves."""
    if isinstance(basis, core.PlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    if isinstance(to_basis, core.PlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kx, ky, kz = to_basis.kvecs(k0, material, modetype)
    res = np.array(
        where
        & (kx[:, None] == kvecs[0])
        & (ky[:, None] == kvecs[1])
        & (kz[:, None] == kvecs[2]),
        int,
    )
    return core.PhysicsArray(
        res, basis=(to_basis, basis), k0=k0, material=material, modetype=modetype
    )


def expand(
    basis, modetype=None, *, k0=None, material=Material(), poltype=None, where=True,
):
    """Expansion matrix.

    Expand the modes from one basis set to another basis set. If applicable the modetype
    can also be changed, like for spherical and cylindrical waves from `singular` to
    `regular`. Not all expansions are available, only those that result in a discrete
    set of modes. For example, plane waves can be expanded in spherical waves, but the
    opposite transformation generally requires a continuous spectrum (an integral) over
    plane waves.

    Args:
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        modetype (str, optional): Wave mode, used for
            :class:`~treams.SphericalWaveBasis` and :class:`CylindricalWaveBasis`.
        k0 (float, optional): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
        where (array-like, bool, optional): Only evaluate parts of the expansion matrix,
            the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    if isinstance(modetype, (tuple, list)):
        to_modetype, modetype = modetype
    else:
        to_modetype = None
    poltype = config.POLTYPE if poltype is None else poltype
    material = Material(material)
    if isinstance(basis, core.SphericalWaveBasis) and isinstance(
        to_basis, core.SphericalWaveBasis
    ):
        modetype = "regular" if modetype is None else modetype
        to_modetype = modetype if to_modetype is None else to_modetype
        return _sw_sw_expand(
            basis, to_basis, to_modetype, k0, material, modetype, poltype, where
        )
    if isinstance(basis, core.CylindricalWaveBasis):
        if isinstance(to_basis, core.CylindricalWaveBasis):
            modetype = "regular" if modetype is None else modetype
            to_modetype = modetype if to_modetype is None else to_modetype
            return _cw_cw_expand(
                basis, to_basis, to_modetype, k0, material, modetype, where
            )
        if isinstance(to_basis, core.SphericalWaveBasis):
            if modetype != "regular" and to_modetype not in (None, "regular"):
                raise ValueError("invalid modetype")
            return _sw_cw_expand(basis, to_basis, k0, material, poltype, where)
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(to_basis, core.PlaneWaveBasis):
            to_modetype = modetype if to_modetype is None else to_modetype
            return _pw_pw_expand(basis, to_basis, k0, material, modetype, where)
        if isinstance(to_basis, core.CylindricalWaveBasis):
            if to_modetype not in (None, "regular"):
                raise ValueError("invalid modetype")
            return _cw_pw_expand(basis, to_basis, k0, material, modetype, where)
        if isinstance(to_basis, core.SphericalWaveBasis):
            if to_modetype not in (None, "regular"):
                raise ValueError("invalid modetype")
            return _sw_pw_expand(
                basis, to_basis, k0, material, modetype, poltype, where
            )
    raise TypeError("invalid basis")


class Expand(Operator):
    """Expansion matrix.

    When called as attribute of an object it returns a suitable transformation matrix to
    expand one set of modes into another basis set (and mode type, if applicable).
    See also :func:`expand`.
    """

    _FUNC = staticmethod(expand)

    def __init__(self, basis, modetype=None, *, isinv=False):
        args = (basis,) if modetype is None else (basis, modetype)
        super().__init__(*args, isinv=isinv)

    def __call__(self, **kwargs):
        if self.isinv:
            return self._call_inv(**kwargs)
        args = list(self._args)
        if "basis" in kwargs and not isinstance(args[0], tuple):
            args[0] = (args[0], kwargs.pop("basis"))
        if "modetype" in kwargs and len(args) > 1 and not isinstance(args[1], tuple):
            args[1] = (args[1], kwargs.pop("modetype"))
        return self.FUNC(*args, **kwargs)

    def _call_inv(self, **kwargs):
        args = list(self._args)
        if "basis" in kwargs and not isinstance(args[0], tuple):
            args[0] = (kwargs.pop("basis"), args[0])
        if "modetype" in kwargs and len(args) > 1 and not isinstance(args[1], tuple):
            args[1] = (kwargs.pop("modetype"), args[1])
        return self.FUNC(*args, **kwargs)

    def get_kwargs(self, obj, dim=-1):
        kwargs = super().get_kwargs(obj, dim)
        for name in ("basis", "modetype"):
            val = getattr(obj, name, None)
            if isinstance(val, tuple):
                val = val[dim]
            if val is not None:
                kwargs[name] = val
        return kwargs

    @property
    def inv(self):
        """Inverse expansion.

        The inverse transformation is not available for all transformations.
        """
        if len(self._args) == 1:
            basis, modetype = self._args[0], None
        else:
            basis, modetype = self._args
        if isinstance(basis, tuple):
            basis = tuple(basis[::-1])
        if isinstance(modetype, tuple):
            modetype = tuple(modetype[::-1])
        return type(self)(basis, modetype, isinv=not self.isinv)


def _swl_expand(basis, to_basis, eta, k0, kpar, lattice, material, poltype, where):
    """Expand spherical waves in a lattice."""
    ks = k0 * material.nmp
    if lattice.dim == 3:
        try:
            length = len(kpar)
        except TypeError:
            length = 1
        if length == 1:
            lattice = Lattice(lattice, "z")
        elif length == 2:
            lattice = Lattice(lattice, "xy")
        elif length == 3:
            # Last attempt to determine the dimension of the sum
            if np.isnan(kpar[2]):
                lattice = Lattice(lattice, "xy")
            elif np.isnan(kpar[0]) and np.isnan(kpar[1]):
                lattice = Lattice(lattice, "z")
    kpar = WaveVector(kpar)
    if lattice.dim == 1:
        x = kpar[2]
    elif lattice.dim == 2:
        x = kpar[:2]
    else:
        x = kpar[:]
    res = sw.translate_periodic(
        ks,
        x,
        lattice[...],
        to_basis.positions,
        to_basis[()],
        basis[()],
        basis.positions,
        poltype=poltype,
        eta=eta,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        poltype=poltype,
        material=material,
        lattice=lattice,
        modetype=("regular", "singular"),
        kpar=kpar,
    )


def _cw_sw_expand(basis, to_basis, k0, kpar, lattice, material, poltype, where):
    """Expand spherical waves in a lattice in cylindrical waves."""
    ks = k0 * material.nmp
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    kpar = WaveVector(kpar)
    res = sw.periodic_to_cw(
        *(m[:, None] for m in to_basis.zms),
        *basis.lms,
        ks[basis.pol],
        Lattice(lattice, "z").volume,
        poltype=poltype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        poltype=poltype,
        material=material,
        modetype="singular",
        lattice=lattice,
        kpar=kpar,
    )


def _pw_sw_expand(
    basis, to_basis, k0, kpar, lattice, material, modetype, poltype, where
):
    """Expand spherical waves in a lattice in plane waves."""
    if modetype is None and isinstance(to_basis, core.PlaneWaveBasisByComp):
        modetype = "up"
    kpar = WaveVector(kpar)
    res = sw.periodic_to_pw(
        *(b[:, None] for b in to_basis.kvecs(k0, material, modetype)),
        to_basis.pol[:, None],
        *basis.lms,
        Lattice(lattice, "xy").volume,
        poltype=poltype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        kpar=kpar,
        lattice=lattice,
        material=material,
        modetype=(modetype, "singular"),
        poltype=poltype,
    )


def _cwl_expand(basis, to_basis, eta, k0, kpar, lattice, material, where):
    """Expand cylindrical waves in a lattice."""
    ks = material.ks(k0)
    alignment = (
        "x" if not isinstance(lattice, Lattice) and np.size(lattice) == 1 else None
    )
    lattice = Lattice(lattice, alignment)
    if lattice.dim == 3:
        try:
            length = len(kpar)
        except TypeError:
            length = 1
        if length == 1:
            lattice = Lattice(lattice, "x")
        elif length == 2:
            lattice = Lattice(lattice, "xy")
        elif length == 3:
            # Last attempt to determine the dimension of the sum
            if np.isnan(kpar[1]):
                lattice = Lattice(lattice, "x")
                kpar = WaveVector(kpar, "x")
            else:
                lattice = Lattice(lattice, "xy")
    if lattice.dim == 1:
        kpar = WaveVector(kpar, "x")
        x = kpar[0]
    elif lattice.dim == 2:
        kpar = WaveVector(kpar)
        x = kpar[:2]
    res = cw.translate_periodic(
        ks,
        x,
        lattice[...],
        to_basis.positions,
        to_basis[()],
        basis[()],
        basis.positions,
        eta=eta,
    )
    res[..., np.logical_not(where)] = 0
    for b in (to_basis, basis):
        if b.kpar is not None:
            kpar = kpar & b.kpar
    return core.PhysicsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        lattice=lattice,
        modetype=("regular", "singular"),
        kpar=kpar,
    )


def _pw_cw_expand(basis, to_basis, k0, lattice, kpar, material, modetype, where):
    """Expand cylindrical waves in a lattice in plane waves."""
    if modetype is None and isinstance(to_basis, core.PlaneWaveBasisByComp):
        modetype = "up"
    if len(kpar) == 1:
        kpar = WaveVector(kpar, alignment="x")
    res = cw.periodic_to_pw(
        *(b[:, None] for b in to_basis.kvecs(k0, material, modetype)),
        to_basis.pol[:, None],
        *basis.zms,
        lattice.volume,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.PhysicsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        modetype=(modetype, "singular"),
        lattice=lattice,
        kpar=kpar,
    )


def expandlattice(
    lattice=None,
    kpar=None,
    basis=None,
    modetype=None,
    *,
    eta=0,
    k0=None,
    material=Material(),
    poltype=None,
    where=True,
):
    """Expansion matrix in lattices.

    Expand the modes from one basis set which are assumed to be periodically repeated on
    a lattice into another basis set.

    Args:
        lattice (:class:`~treams.Lattice` or array-like, optional): Lattice definition.
            In some cases this argument can be omitted, when the lattice can be inferred
            from the basis.
        kpar (sequence, optional): The components of the wave vector tangential to the
            lattice. In some cases this argument can be omitted, when the lattice can be
            inferred from the basis.
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        eta (float or complex, optional): Split parameter used when the Ewald summation
            is applied for the lattice sums. By setting it to 0 the split is set
            automatically.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode, used for
            :class:`~treams.PlaneWaveBasisByComp`.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
        where (array-like, bool, optional): Only evaluate parts of the expansion matrix,
            the give array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    if lattice is None and to_basis.lattice is not None:
        lattice = to_basis.lattice
    if not isinstance(lattice, Lattice) and np.size(lattice) == 1:
        alignment = "x" if isinstance(basis, core.CylindricalWaveBasis) else "z"
    elif isinstance(to_basis, core.PlaneWaveBasis) and isinstance(
        basis, core.CylindricalWaveBasis
    ):
        alignment = "x"
    else:
        alignment = None
    lattice = Lattice(lattice, alignment)
    if kpar is None:
        if basis.kpar is None:
            kpar = to_basis.kpar
        else:
            kpar = basis.kpar
    try:
        kpar = list(kpar)
    except TypeError:
        if kpar is None:
            kpar = [np.nan] * 3
        else:
            kpar = [kpar]
    poltype = config.POLTYPE if poltype is None else poltype
    material = Material(material)
    if isinstance(basis, core.SphericalWaveBasis):
        if isinstance(to_basis, core.SphericalWaveBasis):
            return _swl_expand(
                basis, to_basis, eta, k0, kpar, lattice, material, poltype, where
            )
        if isinstance(to_basis, core.CylindricalWaveBasis):
            return _cw_sw_expand(
                basis, to_basis, k0, kpar, lattice, material, poltype, where
            )
        if isinstance(to_basis, core.PlaneWaveBasis):
            if isinstance(modetype, tuple):
                modetype = modetype[0]
            return _pw_sw_expand(
                basis, to_basis, k0, kpar, lattice, material, modetype, poltype, where
            )
    if isinstance(basis, core.CylindricalWaveBasis):
        if isinstance(to_basis, core.CylindricalWaveBasis):
            return _cwl_expand(basis, to_basis, eta, k0, kpar, lattice, material, where)
        if isinstance(to_basis, core.PlaneWaveBasis):
            if isinstance(modetype, tuple):
                modetype = modetype[0]
            return _pw_cw_expand(
                basis, to_basis, k0, lattice, kpar, material, modetype, where
            )
    raise TypeError("invalid basis")


class ExpandLattice(Operator):
    """Expansion matrix in a lattice.

    When called as attribute of an object it returns a suitable transformation matrix to
    expand one set of modes that is periodically repeated into another basis set.
    See also :func:`expandlattice`.
    """

    _FUNC = staticmethod(expandlattice)

    def __init__(self, lattice=None, kpar=None, basis=None, modetype=None):
        # if lattice is None:
        #     if basis is None:
        #         raise ValueError("'basis' or 'lattice' needs to be defined")
        #     elif basis.lattice is None:
        #         raise ValueError("'basis.lattice' or 'lattice' needs to be defined")
        super().__init__(lattice, kpar, basis, modetype)

    def __call__(self, **kwargs):
        if self.isinv:
            return self._call_inv(**kwargs)
        args = list(self._args)
        if "basis" in kwargs:
            if args[2] is None:
                args[2] = kwargs.pop("basis")
            else:
                args[2] = (args[2], kwargs.pop("basis"))
        if "modetype" in kwargs and args[3] is None:
            args[3] = kwargs.pop("modetype")
        for i, name in enumerate(("lattice", "kpar")):
            if name in kwargs and args[i] is None:
                args[i] = kwargs.pop(name)
        return self.FUNC(*args, **kwargs)

    def get_kwargs(self, obj, dim=-1):
        kwargs = super().get_kwargs(obj, dim)
        kwargs.pop("modetype", None)
        for name in ("basis", "lattice", "kpar"):
            val = getattr(obj, name, None)
            if isinstance(val, tuple):
                val = val[dim]
            if val is not None:
                kwargs[name] = val
        return kwargs

    @property
    def inv(self):
        """Inverse expansion for periodic arrangements.

        The inverse transformation is not available.
        """
        raise NotImplementedError


def _pwp_permute(basis, n, k0, material, modetype, poltype):
    """Permute axes in a partial plane wave basis."""
    if material is None:
        raise TypeError("missing definition of 'material'")
    modetype = "up" if modetype is None else modetype
    kvecs = np.array(basis.kvecs(k0, material, modetype))
    where = (kvecs[..., None] == kvecs[:, None, :]).all(0)
    if n == 1:
        res = pw.permute_xyz(
            kvecs[0],
            kvecs[1],
            kvecs[2],
            basis.pol[:, None],
            basis.pol,
            poltype=poltype,
            where=where,
        )
    elif n == 2:
        res = pw.permute_xyz(
            kvecs[0],
            kvecs[1],
            kvecs[2],
            basis.pol[:, None],
            basis.pol,
            poltype=poltype,
            inverse=True,
            where=where,
        )
    else:
        res = np.eye(len(basis))
    res[np.logical_not(where)] = 0
    return core.PhysicsArray(res, basis=(basis.permute(n), basis), poltype=poltype)


def _pwa_permute(basis, n, poltype):
    """Permute axes in a plane wave basis."""
    qx, qy, qz = basis.qx, basis.qy, basis.qz
    where = (qx[:, None] == qx) & (qy[:, None] == qy) & (qz[:, None] == qz)
    if n == 1:
        res = pw.permute_xyz(
            qx, qy, qz, basis.pol[:, None], basis.pol, poltype=poltype, where=where,
        )
    elif n == 2:
        res = pw.permute_xyz(
            qx,
            qy,
            qz,
            basis.pol[:, None],
            basis.pol,
            poltype=poltype,
            inverse=True,
            where=where,
        )
    elif n == 0:
        res = np.eye(len(basis))
    res[np.logical_not(where)] = 0
    return core.PhysicsArray(res, basis=(basis.permute(n), basis), poltype=poltype)


def permute(n=1, *, basis, k0=None, material=None, modetype=None, poltype=None):
    """Permutation matrix.

    Permute the axes of a plane wave basis expansion.

    Args:
        n (int, optional): Number of permutations, defaults to 1.
        basis (:class:`~treams.BasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    poltype = config.POLTYPE if poltype is None else poltype
    if n != int(n):
        raise ValueError("'n' must be integer")
    n = n % 3
    if isinstance(basis, core.PlaneWaveBasisByComp):
        return _pwp_permute(basis, n, k0, material, modetype, poltype)
    if isinstance(basis, core.PlaneWaveBasisByUnitVector):
        return _pwa_permute(basis, n, poltype)
    raise TypeError("invalid basis")


class Permute(Operator):
    """Axes permutation matrix.

    When called as attribute of an object it returns a suitable transformation matrix to
    permute the axis definitions of plane waves. See also :func:`permute`.
    """

    _FUNC = staticmethod(permute)

    def __init__(self, n=1, *, isinv=False):
        super().__init__(n, isinv=isinv)

    @property
    def inv(self):
        return type(self)(*self._args, isinv=not self.isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs:
            kwargs["basis"] = kwargs["basis"].permute(self._args[0])
        return self.FUNC(-self._args[0], **kwargs)


def _sw_efield(r, basis, k0, material, modetype, poltype):
    """Electric field of spherical waves."""
    ks = k0 * material.nmp
    rsph = sc.car2sph(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = sc.vsw_rA(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                basis.pol,
            )
        elif modetype == "singular":
            res = sc.vsw_A(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                basis.pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = (1 - basis.pol[:, None]) * sc.vsw_rM(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + basis.pol[:, None] * sc.vsw_rN(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
        elif modetype == "singular":
            res = (1 - basis.pol[:, None]) * sc.vsw_M(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + basis.pol[:, None] * sc.vsw_N(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vsph2car(res, rsph[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _cw_efield(r, basis, k0, material, modetype, poltype):
    """Electric field of cylindrical waves."""
    material = Material(material)
    ks = material.ks(k0)[basis.pol]
    krhos = material.krhos(k0, basis.kz, basis.pol)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    poltype = config.POLTYPE if poltype is None else poltype
    rcyl = sc.car2cyl(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = sc.vcw_rA(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
        elif modetype == "singular":
            res = sc.vcw_A(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = (1 - basis.pol[:, None]) * sc.vcw_rM(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + basis.pol[:, None] * sc.vcw_rN(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
        elif modetype == "singular":
            res = (1 - basis.pol[:, None]) * sc.vcw_M(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + basis.pol[:, None] * sc.vcw_N(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vcyl2car(res, rcyl[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _pw_efield(r, basis, k0, material, modetype, poltype):
    """Electric field of plane waves."""
    res = None
    kvecs = basis.kvecs(k0, material, modetype)
    if poltype == "helicity":
        res = sc.vpw_A(*kvecs, r[..., 0], r[..., 1], r[..., 2], basis.pol,)
    elif poltype == "parity":
        res = (1 - basis.pol[:, None]) * sc.vpw_M(
            *kvecs, r[..., 0], r[..., 1], r[..., 2],
        ) + basis.pol[:, None] * sc.vpw_N(*kvecs, r[..., 0], r[..., 1], r[..., 2],)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    return res


def efield(r, *, basis, k0, material=Material(), modetype=None, poltype=None):
    """Electric field.

    The resulting matrix maps the electric field coefficients of the given basis to the
    electric field in Cartesian coordinates.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_efield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_efield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_efield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class FieldOperator(Operator):
    def __init__(self, r):
        super().__init__(r)

    @property
    def inv(self):
        """Inverse transformation of an electric field to modes.

        The inverse transformation is not available.
        """
        raise NotImplementedError


class EField(FieldOperator):
    """Electric field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`efield`.
    """

    _FUNC = staticmethod(efield)


def _sw_hfield(r, basis, k0, material, modetype, poltype):
    """Magnetic field of spherical waves."""
    ks = k0 * material.nmp
    rsph = sc.car2sph(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = (2 * basis.pol[:, None] - 1) * sc.vsw_rA(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                basis.pol,
            )
        elif modetype == "singular":
            res = (2 * basis.pol[:, None] - 1) * sc.vsw_A(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                basis.pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = basis.pol[:, None] * sc.vsw_rM(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + (1 - basis.pol[:, None]) * sc.vsw_rN(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
        elif modetype == "singular":
            res = basis.pol[:, None] * sc.vsw_M(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + (1 - basis.pol[:, None]) * sc.vsw_N(
                basis.l,
                basis.m,
                ks[basis.pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
    res *= -1j / material.impedance
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vsph2car(res, rsph[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _cw_hfield(r, basis, k0, material, modetype, poltype):
    """Magnetic field of cylindrical waves."""
    material = Material(material)
    ks = material.ks(k0)[basis.pol]
    krhos = material.krhos(k0, basis.kz, basis.pol)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    poltype = config.POLTYPE if poltype is None else poltype
    rcyl = sc.car2cyl(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = (2 * basis.pol[:, None] - 1) * sc.vcw_rA(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
        elif modetype == "singular":
            res = (2 * basis.pol[:, None] - 1) * sc.vcw_A(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = basis.pol[:, None] * sc.vcw_rM(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + (1 - basis.pol[:, None]) * sc.vcw_rN(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
        elif modetype == "singular":
            res = basis.pol[:, None] * sc.vcw_M(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + (1 - basis.pol[:, None]) * sc.vcw_N(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
    res *= -1j / material.impedance
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vcyl2car(res, rcyl[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _pw_hfield(r, basis, k0, material, modetype, poltype):
    """Magnetic field of plane waves."""
    res = None
    kvecs = basis.kvecs(k0, material, modetype)
    if poltype == "helicity":
        res = (2 * basis.pol[:, None] - 1) * sc.vpw_A(
            *kvecs, r[..., 0], r[..., 1], r[..., 2], basis.pol,
        )
    elif poltype == "parity":
        res = basis.pol[:, None] * sc.vpw_M(
            *kvecs, r[..., 0], r[..., 1], r[..., 2],
        ) + (1 - basis.pol[:, None]) * sc.vpw_N(
            *kvecs, r[..., 0], r[..., 1], r[..., 2],
        )
    res *= -1j / material.impedance
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    return res


def hfield(r, *, basis, k0, material=Material(), modetype=None, poltype=None):
    r"""Magnetic field.

    The resulting matrix maps the electric field coefficients of the given basis to the
    magnetic field in Cartesian coordinates.

    The magnetic field is given in units of :math:`\frac{1}{Z_0} [\boldsymbol E]`.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_hfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_hfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_hfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class HField(FieldOperator):
    """Magnetic field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`hfield`.
    """

    _FUNC = staticmethod(hfield)


def _sw_dfield(r, basis, k0, material, modetype, poltype):
    """Displacement field of spherical waves."""
    res = _sw_efield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] / material.impedance
    else:
        res *= material.epsilon
    return res


def _cw_dfield(r, basis, k0, material, modetype, poltype):
    """Displacement field of cylindrical waves."""
    res = _cw_efield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] / material.impedance
    else:
        res *= material.epsilon
    return res


def _pw_dfield(r, basis, k0, material, modetype, poltype):
    """Displacement field of partial plane waves."""
    res = _pw_efield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] / material.impedance
    else:
        res *= material.epsilon
    return res


def dfield(r, *, basis, k0, material=Material(), modetype=None, poltype=None):
    r"""Displacement field.

    The resulting matrix maps the electric field coefficients of the given basis to the
    displacement field in Cartesian coordinates.

    The displacement field is given in units of :math:`\epsilon_0 [\boldsymbol E]`.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_dfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_dfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_dfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class DField(FieldOperator):
    """Displacement field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`dfield`.
    """

    _FUNC = staticmethod(dfield)


def _sw_bfield(r, basis, k0, material, modetype, poltype):
    """Magnetic flux density of spherical waves."""
    res = _sw_hfield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] * material.impedance
    else:
        res *= material.mu
    return res


def _cw_bfield(r, basis, k0, material, modetype, poltype):
    """Magnetic flux density of cylindrical waves."""
    res = _cw_hfield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] * material.impedance
    else:
        res *= material.mu
    return res


def _pw_bfield(r, basis, k0, material, modetype, poltype):
    """Magnetic flux density of partial plane waves."""
    res = _pw_hfield(r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol][:, None] * material.impedance
    else:
        res *= material.mu
    return res


def bfield(r, *, basis, k0=None, material=Material(), modetype=None, poltype=None):
    r"""Magnetic flux density.

    The resulting matrix maps the electric field coefficients of the given basis to the
    magnetic flux density in Cartesian coordinates.

    The magnetic flux density is given in units of
    :math:`\frac{1}{c_0} [\boldsymbol E]`.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float, optional): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_bfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_bfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_bfield(r, basis, k0, material, modetype, poltype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class BField(FieldOperator):
    """Magnetic flux density evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`bfield`.
    """

    _FUNC = staticmethod(bfield)


def _sw_gfield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field G of spherical waves."""
    ks = k0 * material.nmp
    rsph = sc.car2sph(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = (basis.pol[:, None] == pol) * sc.vsw_rA(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                pol,
            )
        elif modetype == "singular":
            res = (basis.pol[:, None] == pol) * sc.vsw_A(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
                pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = (1 + 2 * (pol - 1) * basis.pol[:, None]) * sc.vsw_rM(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + (2 * pol - 1 - 2 * (pol - 1) * basis.pol[:, None]) * sc.vsw_rN(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
        elif modetype == "singular":
            res = (1 + 2 * (pol - 1) * basis.pol[:, None]) * sc.vsw_M(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            ) + (2 * pol - 1 - 2 * (pol - 1) * basis.pol[:, None]) * sc.vsw_N(
                basis.l,
                basis.m,
                ks[pol] * rsph[..., basis.pidx, 0],
                rsph[..., basis.pidx, 1],
                rsph[..., basis.pidx, 2],
            )
    if res is None:
        raise ValueError("invalid parameters")
    res *= np.sqrt(2)
    res = util.AnnotatedArray(sc.vsph2car(res, rsph[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _cw_gfield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field G of cylindrical waves."""
    material = Material(material)
    ks = material.ks(k0)[basis.pol]
    krhos = material.krhos(k0, basis.kz, basis.pol)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    poltype = config.POLTYPE if poltype is None else poltype
    rcyl = sc.car2cyl(r - basis.positions)
    res = None
    if poltype == "helicity":
        if modetype == "regular":
            res = (basis.pol[:, None] == pol) * sc.vcw_rA(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
        elif modetype == "singular":
            res = (basis.pol[:, None] == pol) * sc.vcw_A(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
                basis.pol,
            )
    elif poltype == "parity":
        if modetype == "regular":
            res = (1 + 2 * (pol - 1) * basis.pol[:, None]) * sc.vcw_rM(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + (2 * pol - 1 - 2 * (pol - 1) * basis.pol[:, None]) * sc.vcw_rN(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
        elif modetype == "singular":
            res = (1 + 2 * (pol - 1) * basis.pol[:, None]) * sc.vcw_M(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
            ) + (2 * pol - 1 - 2 * (pol - 1) * basis.pol[:, None]) * sc.vcw_N(
                basis.kz,
                basis.m,
                krhos * rcyl[..., basis.pidx, 0],
                rcyl[..., basis.pidx, 1],
                rcyl[..., basis.pidx, 2],
                ks,
            )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vcyl2car(res, rcyl[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["poltype"] = poltype
    res.ann[-2]["modetype"] = modetype
    return res


def _pw_gfield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field G of plane waves."""
    res = None
    kvecs = basis.kvecs(k0, material, modetype)
    if poltype == "helicity":
        res = (basis.pol[:, None] == pol) * sc.vpw_A(
            *kvecs, r[..., 0], r[..., 1], r[..., 2], basis.pol,
        )
    elif poltype == "parity":
        res = (1 + 2 * (pol - 1) * basis.pol[:, None]) * sc.vpw_M(
            *kvecs, r[..., 0], r[..., 1], r[..., 2],
        ) + (2 * pol - 1 - 2 * (pol - 1) * basis.pol[:, None]) * sc.vpw_N(
            *kvecs, r[..., 0], r[..., 1], r[..., 2],
        )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["poltype"] = poltype
    return res


def gfield(pol, r, *, basis, k0, material=Material(), modetype=None, poltype=None):
    """Riemann-Silberstein field G.

    The resulting matrix maps the electric field coefficients of the given basis to the
    Riemann-Silberstein field G in Cartesian coordinates.

    Args:
        pol (int): Type of the Riemann-Silberstein field, must be 1 or -1. In analogy to
            to polarizations for helicity mode, 0 is also treated as -1.
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    if pol == -1:
        pol = 0
    elif pol not in (0, 1):
        raise ValueError(f"invalid argument pol: '{pol}'")
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_gfield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_gfield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_gfield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    raise TypeError("invalid basis")


class GField(FieldOperator):
    """Riemann-Silberstein field G evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`gfield`.
    """

    _FUNC = staticmethod(gfield)


def _sw_ffield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field F of spherical waves."""
    res = _sw_gfield(pol, r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol, None] / material.n
    return res


def _cw_ffield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field F of cylindrical waves."""
    res = _cw_gfield(pol, r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol, None] / material.n
    return res


def _pw_ffield(pol, r, basis, k0, material, modetype, poltype):
    """Riemann-Silberstein field F of partial plane waves."""
    res = _pw_gfield(pol, r, basis, k0, material, modetype, poltype)
    if poltype == "helicity":
        res *= material.nmp[basis.pol, None] / material.n
    return res


def ffield(pol, r, *, basis, k0, material=Material(), modetype=None, poltype=None):
    """Riemann-Silberstein field F.

    The resulting matrix maps the electric field coefficients of the given basis to the
    Riemann-Silberstein field F in Cartesian coordinates.

    Args:
        pol (int): Type of the Riemann-Silberstein field, must be 1 or -1. In analogy to
            to polarizations for helicity mode, 0 is also treated as -1.
        r (array-like): Evaluation points
        basis (:class:`~treams.BasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~treams.Material` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
        poltype (str, optional): Polarization, see also
            :ref:`params:Polarizations`.
    """
    if pol == -1:
        pol = 0
    elif pol not in (0, 1):
        raise ValueError(f"invalid argument pol: '{pol}'")
    material = Material(material)
    poltype = config.POLTYPE if poltype is None else poltype
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.SphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _sw_ffield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    if isinstance(basis, core.CylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _cw_ffield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    if isinstance(basis, core.PlaneWaveBasis):
        if isinstance(basis, core.PlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _pw_ffield(pol, r, basis, k0, material, modetype, poltype).swapaxes(
            -1, -2
        )
    raise TypeError("invalid basis")


class FField(FieldOperator):
    """Riemann-Silberstein field F evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`ffield`.
    """

    _FUNC = staticmethod(ffield)



import matplotlib.pyplot as plt
import numpy as np

import treams

k0s = 2 * np.pi * np.linspace(0.01, 1, 200)
materials = [treams.Material(12.4), treams.Material()]
mmax = 3
radius = 0.2
lattice = treams.Lattice.square(0.3)
kpar = [0, 0, 0]

res = []
for k0 in k0s:
    cyl = treams.TMatrixC.cylinder(kpar[2], mmax, k0, radius, materials)
    svd = np.linalg.svd(cyl.latticeinteraction(lattice, kpar[:2]), compute_uv=False)
    res.append(svd[-1])

fig, ax = plt.subplots()
ax.set_xlabel("Frequency (THz)")
ax.set_ylabel("Smallest singular value")
ax.semilogy(299.792458 * k0s / (2 * np.pi), res)
fig.show()



"""Utilities.

Collection of functions to make :class:`AnnotatedArray` work. The implementation is
aimed to be quite general, while still providing a solid basis for the rest of the code.
"""

import abc
import collections
import copy
import itertools
import warnings

import numpy as np

import treams._operators as _op

__all__ = [
    "AnnotatedArray",
    "AnnotationDict",
    "AnnotationError",
    "AnnotationSequence",
    "AnnotationWarning",
    "implements",
    "OrderedSet",
]


class OrderedSet(collections.abc.Sequence, collections.abc.Set):
    """Ordered set.

    A abstract base class that combines a sequence and set. In contrast to regular sets
    it is expected that the equality comparison only returns `True` if all entries are
    in the same order.
    """

    @abc.abstractmethod
    def __eq__(self, other):
        """Equality test."""
        raise NotImplementedError


class AnnotationWarning(UserWarning):
    """Custom warning for Annotations.

    By default the warning filter is set to 'always'.
    """


warnings.simplefilter("always", AnnotationWarning)


class AnnotationError(Exception):
    """Custom exception for Annotations."""


class AnnotationDict(collections.abc.MutableMapping):
    """Dictionary that notifies the user when overwriting keys.

    Behaves mostly similar to regular dictionaries, except that when overwriting
    existing keys a :class:`AnnotationWarning` is emmitted.

    Examples:
        An :class:`AnnotationDict` can be created from other mappings, from a list of
        key-value pairs (note how a warning is emmitted for the duplicate key), and from
        keyword arguments.

        .. code-block:: python

            >>> AnnotationDict({"a": 1, "b": 2})
            AnnotationDict({'a': 1, 'b': 2})
            >>> AnnotationDict([("a", 1), ("b", 2), ("a", 3)])
            treams/util.py:74: AnnotationWarning: overwriting key 'a'
            warnings.warn(f"overwriting key '{key}'", AnnotationWarning)
            AnnotationDict({'a': 3, 'b': 2})
            >>> AnnotationDict({"a": 1, "b": 2}, c=3)
            AnnotationDict({'a': 1, 'b': 2, 'c': 3})

    Warns:
        AnnotationWarning
    """

    def __init__(self, items=(), /, **kwargs):
        """Initialization."""
        self._dct = {}
        for i in (items.items() if hasattr(items, "items") else items, kwargs.items()):
            for key, val in i:
                self[key] = val

    def __getitem__(self, key):
        """Get a value by its key.

        Args:
            key (hashable): Key
        """
        return self._dct[key]

    def __setitem__(self, key, val):
        """Set item specified by key to the defined value.

        When overwriting an existing key an :class:`AnnotationWarning` is emitted.
        Avoid the warning by explicitly deleting the key first.

        Args:
            key (hashable): Key
            val : Value

        Warns:
            AnnotationWarning
        """
        if key in self and self[key] != val:
            warnings.warn(f"overwriting key '{key}'", AnnotationWarning)
        self._dct[key] = val

    def __delitem__(self, key):
        """Delete the key."""
        del self._dct[key]

    def __iter__(self):
        """Iterate over the keys.

        Returns:
            Iterator
        """
        return iter(self._dct)

    def __len__(self):
        """Number of keys contained.

        Returns:
            int
        """
        return len(self._dct)

    def __repr__(self):
        """String representation.

        Returns:
            str
        """
        return f"{self.__class__.__name__}({repr(self._dct)})"

    def match(self, other):
        """Compare the own keys to another dictionary.

        This emits an :class:`AnnotationWarning` for each key that would be overwritten
        by the given dictionary.

        Args:
            other (Mapping)

        Returns:
            None

        Warns:
            AnnotationWarning
        """
        for key, val in self.items():
            if key in other and other[key] != val:
                warnings.warn(f"incompatible key '{key}'", AnnotationWarning)


class SequenceAsDict(collections.abc.MutableMapping):
    def __get__(self, obj, objtype=None):
        self._obj = obj
        return self

    def __set__(self, obj, dct):
        self._obj = obj
        for key, val in dct.items():
            self[key] = val

    def __getitem__(self, key):
        res = tuple(i.get(key) for i in self._obj)
        if all(i is None for i in res):
            raise KeyError(key)
        return res

    def __setitem__(self, key, val):
        val = (None,) * (len(self._obj) - len(val)) + val
        for dct, value in zip(reversed(self._obj), reversed(val)):
            if value is None:
                dct.pop(key, None)
            else:
                dct[key] = value

    def __delitem__(self, key):
        found = False
        for dct in self._obj:
            if key in dct:
                del dct[key]
                found = True
        if not found:
            raise KeyError(key)

    def __iter__(self):
        return iter({key for dct in self._obj for key in dct})

    def __len__(self):
        return len({key for dct in self._obj for key in dct})

    def __repr__(self):
        return f"{self.__class__.__name__}({dict(i for i in self.items())})"


class AnnotationSequence(collections.abc.Sequence):
    """A Sequence of dictionaries.

    This class is intended to work together with :class:`AnnotationDict`. It provides
    convenience functions to interact with multiple of those dictionaries, which are
    mainly used to keep track of the annotations made to each dimension of an
    :class:`AnnotatedArray`. While the sequence itself is immutable the entries of each
    dictionary is mutable.

    Args:
        *args: Items of the sequence
        mapping (AnnotationDict): Type of mapping to use in the sequence.

    Warns:
        AnnotationWarning
    """

    as_dict = SequenceAsDict()

    def __init__(self, *args, mapping=AnnotationDict):
        """Initialization."""
        self._ann = tuple(mapping(i) for i in args)

    def __len__(self):
        """Number of dictionaries in the sequence.

        Returns:
            int
        """
        return len(self._ann)

    def __getitem__(self, key):
        """Get an item or subsequence.

        Indexing works with integers and slices like regular tuples. Additionally, it is
        possible to get a copy of the object with `()`, or a new sequence of mappings in
        a list (or other iterable) of integers.

        Args:
            key (iterable, slice, int)

        Returns:
            mapping
        """
        if isinstance(key, tuple) and key == ():
            return copy.copy(self._ann)
        if isinstance(key, slice):
            return type(self)(*self._ann[key])
        if isinstance(key, (int, np.integer)):
            return self._ann[key]
        res = []
        for k in key:
            if not isinstance(k, int):
                raise TypeError(
                    "sequence index must be integer, slice, list of integers, or '()'"
                )
            res.append(self[k])
        return type(self)(*res)

    def update(self, other):
        """Update all mappings in the sequence at once.

        The given sequence is aliged at the last entry and then updated pairwise.
        Warnings for overwritten keys are extended by the information at which index
        they occurred.

        Args:
            other (Sequence[Mapping]): Mappings to update with.

        Warns:
            AnnotationWarning
        """
        if len(other) > len(self):
            warnings.warn(
                f"argument of length {len(self)} given: "
                f"ignore leading {len(other) - len(self)} entries",
                AnnotationWarning,
            )
        for i, (dest, src) in enumerate(zip(reversed(self), reversed(other))):
            with warnings.catch_warnings():
                warnings.simplefilter("error", category=AnnotationWarning)
                try:
                    dest.update(src)
                except AnnotationWarning as err:
                    warnings.simplefilter("always", category=AnnotationWarning)
                    warnings.warn(
                        f"at index {len(self) - i - 1}: " + err.args[0],
                        AnnotationWarning,
                    )

    def match(self, other):
        """Match all mappings at once.

        The given sequence is aliged at the last entry and then updated pairwise.
        Warnings for overwritten keys are extended by the information at which index
        they occurred, see also :func:`AnnotationDict.match`.

        Args:
            other (Sequence[Mappings]): Mappings to match.

        Warns:
            AnnotationWarning
        """
        for i, (dest, src) in enumerate(zip(reversed(self), reversed(other))):
            with warnings.catch_warnings():
                warnings.simplefilter("error", category=AnnotationWarning)
                try:
                    dest.match(src)
                except AnnotationWarning as err:
                    warnings.simplefilter("always", category=AnnotationWarning)
                    warnings.warn(
                        f"at dimension {len(self) + i}: " + err.args[0],
                        AnnotationWarning,
                    )

    def __eq__(self, other):
        """Equality test.

        Two sequences of mappings are considered equal when they have equal length and
        when they all mappings are equal.

        Args:
            other (Sequence[Mapping]): Mappings to compare with.

        Returns:
            bool
        """
        try:
            lenother = len(other)
        except TypeError:
            return False
        if len(self) != lenother:
            return False
        for a, b in zip(self, other):
            if a != b:
                return False
        return True

    def __repr__(self):
        """String representation.

        Returns:
            str
        """
        if len(self) == 1:
            return f"{type(self).__name__}({repr(self._ann[0])})"
        return f"{type(self).__name__}{repr(self._ann)}"

    def __add__(self, other):
        return AnnotationSequence(*self, *other)

    def __radd__(self, other):
        return AnnotationSequence(*other, *self)


def _cast_nparray(arr):
    """Cast AnnotatedArray to numpy array."""
    return np.asarray(arr) if isinstance(arr, AnnotatedArray) else arr


def _cast_annarray(arr):
    """Cast array to AnnotatedArray."""
    return arr if isinstance(arr, np.generic) else AnnotatedArray(arr)


def _parse_signature(signature, inputdims):
    """Parse a ufunc signature based on the actual inputs.

    The signature is matched with the input dimensions, to get the actual signature. It
    is returned as two lists (inputs and outputs). Each list contains for each
    individual input and output another list of the signature items.

    Example:
        >>> treams.util._parse_signature('(n?,k),(k,m?)->(n?,m?)', (2, 1))
        ([['n?', 'k'], ['k']], [['n?']])

    Args:
        signature (str): Function signature
        inputdims (Iterable[int]): Input dimensions

    Returns:
        tuple[list[list[str]]
    """
    signature = "".join(signature.split())  # remove whitespace
    sigin, sigout = signature.split("->")  # split input and output
    sigin = sigin[1:-1].split("),(")  # split input
    sigin = [i.split(",") for i in sigin]
    sigout = sigout[1:-1].split("),(")  # split output
    sigout = [i.split(",") for i in sigout]
    for i, idim in enumerate(inputdims):
        j = 0
        while j < len(sigin[i]):
            d = sigin[i][j]
            if d.endswith("?") and len(sigin[i]) > idim:
                sigin = [[i for i in s if i != d] for s in sigin]
                sigout = [[i for i in s if i != d] for s in sigout]
            else:
                j += 1
    return sigin, sigout


def _parse_key(key, ndim):
    """Parse a key to index an array.

    This function attempts to replicate the numpy indexing of arrays. It can handle
    integers, slices, an Ellipsis, arrays of integers, arrays of bools, (bare) bools,
    and None. It returns the key with an Ellipsis appended if the number of index
    dimensions does not match the number of dimensions given. Additionally, it informs
    about the number of dimension indexed by fancy indexing, if the fancy indexed
    dimensions will be prepended, and the number of dimensions that the Ellipsis
    contains.

    Args:
        key (tuple): The indexing key
        ndim (int): Number of array dimensions

    Returns:
        tuple
    """
    consumed = 0
    ellipsis = False
    fancy_ndim = 0
    # nfancy = 0
    consecutive_intfancy = 0
    # consecutive_intfancy = 0: no fancy/integer indexing
    # consecutive_intfancy = 1: ongoing consecutive fancy/integer indexing
    # consecutive_intfancy = 2: terminated consecutive fancy/integer indexing
    # consecutive_intfancy >= 2: non-consecutive fancy/integer indexing

    # The first pass gets the consumed dimensions, the presence of an ellipsis, and
    # fancy index properties.
    for k in key:
        if k is not True and k is not False and isinstance(k, (int, np.integer)):
            consumed += 1
            consecutive_intfancy += (consecutive_intfancy + 1) % 2
        elif k is None:
            consecutive_intfancy += consecutive_intfancy % 2
        elif isinstance(k, slice):
            consumed += 1
            consecutive_intfancy += consecutive_intfancy % 2
        elif k is Ellipsis:
            ellipsis = True
            # consumed is determined at the end
            consecutive_intfancy += consecutive_intfancy % 2
        else:
            arr = np.asanyarray(k)
            if arr.dtype == bool:
                consumed += arr.ndim
                fancy_ndim = max(1, fancy_ndim)
            else:
                consumed += 1
                fancy_ndim = max(arr.ndim, fancy_ndim)
            # nfancy += arr.ndim
            consecutive_intfancy += (consecutive_intfancy + 1) % 2

    lenellipsis = ndim - consumed
    if lenellipsis != 0 and not ellipsis:
        key = key + (Ellipsis,)

    return key, fancy_ndim, consecutive_intfancy >= 2, lenellipsis


HANDLED_FUNCTIONS = {}
"""Dictionary of numpy functions implemented for AnnotatedArrays."""


def implements(np_func):
    """Decorator to register an __array_function__ implementation to AnnotatedArrays."""

    def decorator(func):
        HANDLED_FUNCTIONS[np_func] = func
        return func

    return decorator


class AnnotatedArray(np.lib.mixins.NDArrayOperatorsMixin):
    """Array that keeps track of annotations for each dimension.

    This class acts mostly like numpy arrays, but it is enhanced by the following
    functionalities:

        * Annotations are added to each dimension
        * Annotations are compared and preserved for numpy (generalized)
          :py:class:`numpy.ufunc` (like :py:data:`numpy.add`, :py:data:`numpy.exp`,
          :py:data:`numpy.matmul` and many more "standard" mathematical functions)
        * Special ufunc methods, like :py:meth:`numpy.ufunc.reduce`, are supported
        * A growing subset of other numpy functions are supported, like
          :py:func:`numpy.linalg.solve`
        * Keywords can be specified as scale are also index into, when index the
          AnnotatedArray
        * Annotations can also be exposed as properties

    .. testsetup::

        from treams.util import AnnotatedArray

    Example:
        >>> a = AnnotatedArray([[0, 1], [2, 3]], ({"a": 1}, {"b": 2}))
        >>> b = AnnotatedArray([1, 2], ({"b": 2},))
        >>> a @ b
        AnnotatedArray(
            [2, 8],
            AnnotationSequence(AnnotationDict({'a': 1})),
        )

    The interoperability with numpy is implemented using :ref:`basics.dispatch`
    by defining :meth:`__array__`, :meth:`__array_ufunc__`, and
    :meth:`__array_function__`.
    """

    _scales = set()

    def __init__(self, array, ann=(), /, **kwargs):
        """Initalization."""
        self._array = np.asarray(array)
        self.ann = getattr(array, "ann", ())
        self.ann.update(ann)
        for key, val in kwargs.items():
            val = (val,) * self.ndim if not isinstance(val, tuple) else val
            self.ann.as_dict[key] = val

    @classmethod
    def relax(cls, *args, mro=None, **kwargs):
        """Try creating AnnotatedArray subclasses if possible.

        Subclasses can impose stricter conditions on the Annotations. To allow a simple
        "decaying" of those subclasses it is possible to create them with this
        classmethod. It attempts array creations along the method resolution order until
        it succeeds.

        Args:
            mro (array-like, optional): Method resolution order along which to create
                the subclass. By default it takes the order of the calling class.

        Note:
            All other arguments are the same as for the default initialization.
        """
        mro = cls.__mro__[1:] if mro is None else mro
        try:
            return cls(*args, **kwargs)
        except AnnotationError as err:
            if cls == AnnotatedArray:
                raise err from None
        cls, *mro = mro
        return cls.relax(*args, mro=mro, **kwargs)

    def __str__(self):
        """String of the array itself."""
        return str(self._array)

    def __repr__(self):
        """String representation."""
        repr_arr = "    " + repr(self._array)[6:-1].replace("\n  ", "\n")
        return f"{self.__class__.__name__}(\n{repr_arr},\n    {self.ann},\n)"

    def __int__(self):
        return int(self._array)

    def __float__(self):
        return float(self._array)

    def __array__(self, dtype=None):
        """Convert to an numpy array.

        This function returns the bare array without annotations. This function does not
        necessarily make a copy of the array.

        Args:
            dtype (optional): Type of the returned array.
        """
        return np.asarray(self._array, dtype=dtype)

    @property
    def ann(self):
        """Array annotations."""
        return self._ann

    @ann.setter
    def ann(self, ann):
        """Set array annotations.

        This function copies the given sequence of dictionaries.
        """
        self._ann = AnnotationSequence(*(({},) * self.ndim))
        self._ann.update(ann)

    def __getattr__(self, key):
        if key in self.ann.as_dict:
            res = self.ann.as_dict[key]
            if all(res[0] == i for i in res[1:]):
                return res[0]
            return res
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{key}'"
        )

    def __setattr__(self, key, val):
        if key not in ("_array", "ann", "_ann") and key in self.ann.as_dict:
            val = (val,) * self.ndim if not isinstance(val, tuple) else val
            self.ann.as_dict[key] = val
        else:
            super().__setattr__(key, val)

    def __delattr__(self, key):
        try:
            super().__delattr__(key)
        except AttributeError:
            del self.ann.as_dict[key]

    def __len__(self):
        return len(self._array)

    def __copy__(self):
        return type(self)(copy.copy(self._array), copy.copy(self._ann))

    def __deepcopy__(self, memo):
        cls = type(self)
        return cls(copy.deepcopy(self._array, memo), copy.deepcopy(self._ann, memo))

    def __bool__(self):
        """Boolean value of the array."""
        return bool(self._array)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Implement ufunc API."""
        # Compute result first to use numpy's comprehensive checks on the arguments
        inputs_noaa = tuple(map(_cast_nparray, inputs))

        ann_out = ()
        out = kwargs.get("out")
        out = out if isinstance(out, tuple) else (out,)
        ann_out = tuple(getattr(i, "ann", None) for i in out)
        if len(out) != 1 or out[0] is not None:
            kwargs["out"] = tuple(map(_cast_nparray, out))

        res = getattr(ufunc, method)(*inputs_noaa, **kwargs)
        istuple, res = (True, res) if isinstance(res, tuple) else (False, (res,))
        res = tuple(_cast_annarray(r) if o is None else o for r, o in zip(res, out))
        for a, r in zip(ann_out, res):
            if a is not None:
                r.ann.update(a)

        inputs_and_where = inputs + ((kwargs["where"],) if "where" in kwargs else ())

        if ufunc.signature is None or all(
            map(lambda x: x in " (),->", ufunc.signature)
        ):
            if method in ("__call__", "reduceat", "accumulate") or (
                method == "reduce" and kwargs.get("keepdims", False)
            ):
                self._ufunc_call(inputs_and_where, res)
            elif method == "reduce":
                self._ufunc_reduce(inputs_and_where, res, kwargs.get("axis", 0))
            elif method == "at":
                self._ufunc_at(inputs)
                return None
            elif method == "outer":
                self._ufunc_outer(inputs, res, kwargs.get("where", True))
            else:
                warnings.warn("unrecognized ufunc method", AnnotationWarning)
        else:
            res = self._gufunc_call(ufunc, inputs, kwargs, res)
        res = tuple(r if isinstance(r, np.generic) else self.relax(r) for r in res)
        return res if istuple else res[0]

    @staticmethod
    def _ufunc_call(inputs_and_where, res):
        for out, in_ in itertools.product(res, inputs_and_where):
            if not isinstance(out, AnnotatedArray):
                continue
            out.ann.update(getattr(in_, "ann", ()))

    @staticmethod
    def _ufunc_reduce(inputs_and_where, res, axis=0, keepdims=False):
        out = res[0]  # reduce only allowed for single output functions
        if not isinstance(out, AnnotatedArray):
            return
        axis = axis if isinstance(axis, tuple) else (axis,)
        in_ = inputs_and_where[0]
        axis = sorted(map(lambda x: x % np.ndim(in_) - np.ndim(in_), axis))
        for in_ in inputs_and_where:
            ann = list(getattr(in_, "ann", []))
            if not keepdims:
                for a in axis:
                    try:
                        del ann[a]
                    except IndexError:
                        pass
            out.ann.update(ann)

    @staticmethod
    def _ufunc_at(inputs):
        out = inputs[0]
        if not isinstance(out, AnnotatedArray):
            return
        if any(d != {} for d in getattr(inputs[1], "ann", ())):
            warnings.warn("annotations in indices are ignored", AnnotationWarning)
        for in_ in inputs[2:]:
            ann = getattr(in_, "ann", ())
            out.ann.update(ann)

    @staticmethod
    def _ufunc_outer(inputs, res, where=True):
        for out in res:
            if not isinstance(out, AnnotatedArray):
                continue
            in_ann = tuple(
                i for a in inputs for i in getattr(a, "ann", np.ndim(a) * ({},))
            )
            out.ann.update(in_ann)
            where_ann = getattr(where, "ann", ())
            out.ann.update(where_ann)

    @staticmethod
    def _gufunc_call(ufunc, inputs, kwargs, res):
        sigin, sigout = _parse_signature(ufunc.signature, map(np.ndim, inputs))
        if kwargs.get("keepdims", False):
            sigout = [sigin[0] for _ in range(ufunc.nout)]
        ndims = [np.ndim(i) for x in (inputs, res) for i in x]
        axes = getattr(kwargs, "axes", None)
        if axes is None:
            axis = getattr(kwargs, "axis", None)
            if axis is None:
                axes = [tuple(range(-len(i), 0)) for i in sigin + sigout]
            else:
                axes = [(axis,) for _ in range(ufunc.nin)]
        else:
            axes = [
                tuple(a) if isinstance(a, collections.abc.Iterable) else (a,)
                for a in axes
            ]
        append = axes[0] if kwargs.get("keepdims", False) else ()
        axes += [append] * (ufunc.nin + ufunc.nout - len(axes))
        axes = [(*(i % ndim - ndim for i in a),) for a, ndim in zip(axes, ndims)]
        iterdims = [
            [i for i in range(-1, -1 - ndim, -1) if i not in a]
            for a, ndim in zip(axes, ndims)
        ]

        # compare core dimensions
        coredims = {}
        for i, (ax, sig) in enumerate(zip(axes, sigin + sigout)):
            for a, key in zip(ax, sig):
                if key.isnumeric():
                    continue
                coredims.setdefault(key, []).append((i, a))
        inout = tuple(inputs) + res
        for val in coredims.values():
            for (isrc, dimsrc), (idest, dimdest) in itertools.combinations(val, 2):
                source = getattr(inout[isrc], "ann", {dimsrc: {}})[dimsrc]
                dest = getattr(inout[idest], "ann", {dimdest: AnnotationDict()})[
                    dimdest
                ]
                if isrc < ufunc.nin <= idest:
                    dest.update(source)
                else:
                    dest.match(source)

        # compare iteration dimensions
        for iout, out in enumerate(res):
            if not isinstance(out, AnnotatedArray):
                continue
            for idim, dim in enumerate(iterdims[ufunc.nin + iout]):
                dest = out.ann[dim]
                for in_, iterdim in zip(inputs, iterdims):
                    if idim >= len(iterdim) or getattr(in_, "ann", None) is None:
                        continue
                    source = getattr(in_, "ann", {iterdim[idim]: {}})[iterdim[idim]]
                    dest.update(source)
        return res

    def __array_function__(self, func, types, args, kwargs):
        """Function calls on the array.

        Calls defined function in :data:`HANDLED_FUNCTIONS` otherwise raises exception.
        Add functions to it by using the decorator :func:`implements` for custom
        implementations.
        """
        if func not in HANDLED_FUNCTIONS:
            return NotImplemented
        # if not all(issubclass(t, self.__class__) for t in types):
        #     return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __getitem__(self, key):
        """Get an item from the AnnotatedArray.

        The indexing supports most of numpys regular and fancy indexing.
        """
        res = AnnotatedArray(self._array[key])
        if isinstance(res, np.generic) or res.ndim == 0:
            return res

        key = key if isinstance(key, tuple) else (key,)
        key, fancy_ndim, prepend_fancy, lenellipsis = _parse_key(key, self.ndim)
        source = 0
        dest = fancy_ndim if prepend_fancy else 0
        for k in key:
            if k is not True and k is not False and isinstance(k, (int, np.integer)):
                source += 1
            elif k is None:
                dest += 1
            elif isinstance(k, slice):
                for kk, val in self.ann[source].items():
                    if kk in self._scales:
                        res.ann[dest][kk] = val[k]
                    else:
                        res.ann[dest][kk] = val
                dest += 1
                source += 1
            elif k is Ellipsis:
                for _ in range(lenellipsis):
                    res.ann[dest].update(self.ann[source])
                    dest += 1
                    source += 1
            else:
                k = np.asanyarray(k)
                ksq = k.squeeze()
                if ksq.ndim == 1:
                    pos = (np.array(k.shape) == k.size).argmax()
                    ann = (
                        self.ann[source + pos] if k.dtype == bool else self.ann[source]
                    )
                    pos += int(not prepend_fancy) * dest + fancy_ndim - k.ndim
                    for kk, val in ann.items():
                        if kk in self._scales:
                            res.ann[pos][kk] = val[ksq]
                        else:
                            res.ann[pos][kk] = val
                source += k.ndim if k.dtype == bool else 1
                if not prepend_fancy:
                    dest += fancy_ndim
                    fancy_ndim = 0
        return self.relax(res)

    def __setitem__(self, key, value):
        """Set values.

        If the provided value is an AnnotatedArray the annotations of corresponding
        dimensions will be matched.
        """
        self._array[key] = value
        if not hasattr(value, "ann"):
            return
        key = key if isinstance(key, tuple) else (key,)
        key, fancy_ndim, prepend_fancy, lenellipsis = _parse_key(key, self.ndim)
        source = 0
        dest = fancy_ndim if prepend_fancy else 0
        for k in key:
            if k is not True and k is not False and isinstance(k, (int, np.integer)):
                source += 1
            elif k is None:
                dest += 1
            elif isinstance(k, slice):
                for kk, val in self.ann[source].items():
                    if kk not in value.ann[dest]:
                        continue
                    if kk in self._scales:
                        val = val[k]
                    if val != value.ann[dest][kk]:
                        warnings.warn(
                            f"incompatible annotations with key '{kk}' "
                            f"comparing dimensions '{source}' and '{dest}'",
                            AnnotationWarning,
                        )
                dest += 1
                source += 1
            elif k is Ellipsis:
                for _ in range(lenellipsis):
                    self.ann[source].match(value.ann[dest])
                    dest += 1
                    source += 1
            else:
                k = np.asanyarray(k)
                ksq = k.squeeze()
                if ksq.ndim == 1:
                    pos = (np.array(k.shape) == k.size).argmax()
                    ann = (
                        self.ann[source + pos] if k.dtype == bool else self.ann[source]
                    )
                    pos += int(not prepend_fancy) * dest + fancy_ndim - k.ndim
                    for kk, val in ann.items():
                        if kk not in value.ann[pos]:
                            continue
                        if kk in self._scales:
                            val = val[k]
                        if val != value.ann[pos][kk]:
                            warnings.warn(
                                f"incompatible annotations with key '{kk}' "
                                f"comparing dimensions '{source}' and '{pos}'",
                                AnnotationWarning,
                            )
                source += k.ndim if k.dtype == bool else 1
                if not prepend_fancy:
                    dest += fancy_ndim
                    fancy_ndim = 0

    @property
    def T(self):
        """Transpose.

        See also :py:attr:`numpy.ndarray.T`.
        """
        return self.transpose()

    @implements(np.all)
    def all(self, axis=None, dtype=None, out=None, keepdims=False, *, where=True):
        """Test if all elements (along an axis) are True.

        See also :py:meth:`numpy.ndarray.all`.
        """
        return np.logical_and.reduce(self, axis, dtype, out, keepdims, where=where)

    @implements(np.any)
    def any(self, axis=None, dtype=None, out=None, keepdims=False, *, where=True):
        """Test if any element (along an axis) is True.

        See also :py:meth:`numpy.ndarray.any`.
        """
        return np.logical_or.reduce(self, axis, dtype, out, keepdims, where=where)

    @implements(np.max)
    def max(self, axis=None, out=None, keepdims=False, initial=None, where=True):
        """Maximum (along an axis).

        See also :py:meth:`numpy.ndarray.max`.
        """
        return np.maximum.reduce(self, axis, None, out, keepdims, initial, where)

    @implements(np.min)
    def min(self, axis=None, out=None, keepdims=False, initial=None, where=True):
        """Minimum (along an axis).

        See also :py:meth:`numpy.ndarray.min`.
        """
        return np.minimum.reduce(self, axis, None, out, keepdims, initial, where)

    @implements(np.sum)
    def sum(
        self, axis=None, dtype=None, out=None, keepdims=False, initial=0, where=True
    ):
        """Sum of elements (along an axis).

        See also :py:meth:`numpy.ndarray.sum`.
        """
        return np.add.reduce(self, axis, dtype, out, keepdims, initial, where)

    @implements(np.prod)
    def prod(
        self, axis=None, dtype=None, out=None, keepdims=False, initial=1, where=True
    ):
        """Product of elements (along an axis).

        See also :py:meth:`numpy.ndarray.prod`.
        """
        return np.multiply.reduce(self, axis, dtype, out, keepdims, initial, where)

    @implements(np.cumsum)
    def cumsum(self, axis=None, dtype=None, out=None):
        """Cumulative sum of elements (along an axis).

        See also :py:meth:`numpy.ndarray.cumsum`.
        """
        if axis is None:
            return np.add.accumulate(self.flatten(), 0, dtype, out)
        return np.add.accumulate(self, axis, dtype, out)

    @implements(np.cumprod)
    def cumprod(self, axis=None, dtype=None, out=None):
        """Cumulative product of elements (along an axis).

        See also :py:meth:`numpy.ndarray.cumprod`.
        """
        if axis is None:
            return np.multiply.accumulate(self.flatten(), 0, dtype, out)
        return np.multiply.accumulate(self, axis, dtype, out)

    def flatten(self, order="C"):
        """Flatten array to one dimension.

        See also :py:meth:`numpy.ndarray.flatten`.
        """
        res = self.relax(self._array.flatten(order))
        if res.shape == self.shape:
            res.ann.update(self.ann)
        elif len(tuple(filter(lambda x: x != 1, self.shape))) == 1:
            res.ann[0].update(self.ann[self.shape.index(res.size)])
        return res

    @implements(np.trace)
    def trace(self, offset=0, axis1=0, axis2=1, dtype=None, out=None):
        """Trace of an array.

        See also :py:meth:`numpy.ndarray.trac`.
        """
        ann = tuple(a for i, a in enumerate(self.ann) if i not in (axis1, axis2))
        return self.relax(self._array.trace(offset, axis1, axis2, dtype, out), ann)

    def astype(self, *args, **kwargs):
        """Return array as given type.

        See also :py:meth:`numpy.ndarray.astype`.
        """
        return self.relax(self._array.astype(*args, **kwargs), self.ann)

    @property
    @implements(np.ndim)
    def ndim(self):
        """Number of array dimensions.

        See also :py:attr:`numpy.ndarray.ndim`.
        """
        return self._array.ndim

    @property
    @implements(np.shape)
    def shape(self):
        """Array shape.

        See also :py:attr:`numpy.ndarray.shape`.
        """
        return self._array.shape

    @property
    @implements(np.size)
    def size(self):
        """Array size.

        See also :py:attr:`numpy.ndarray.size`.
        """
        return self._array.size

    @property
    @implements(np.imag)
    def imag(self):
        """Imaginary part of the array.

        See also :py:attr:`numpy.ndarray.imag`.
        """
        return self.relax(self._array.imag, self.ann)

    @property
    @implements(np.real)
    def real(self):
        """Real part of the array.

        See also :py:attr:`numpy.ndarray.real`.
        """
        return self.relax(self._array.real, self.ann)

    def conjugate(self, *args, **kwargs):
        """Complex conjugate elementwise.

        See also :py:meth:`numpy.ndarray.conjugate`.
        """
        return np.conjugate(self, *args, **kwargs)

    @implements(np.diagonal)
    def diagonal(self, offset=0, axis1=0, axis2=1):
        """Get the diagonal of the array.

        See also :py:meth:`numpy.ndarray.diagonal`.
        """
        ann = tuple(a for i, a in enumerate(self.ann) if i not in (axis1, axis2)) + (
            {},
        )
        return self.relax(self._array.diagonal(offset, axis1, axis2), ann)

    @implements(np.transpose)
    def transpose(self, axes=None):
        """Transpose array dimensions.

        See also :py:meth:`numpy.ndarray.transpose`.
        """
        axes = range(self.ndim - 1, -1, -1) if axes is None else axes
        return self.relax(self._array.transpose(axes), self.ann[axes])

    conj = conjugate

    @implements(np.swapaxes)
    def swapaxes(self, axis1, axis2):
        axis1 = axis1 % self.ndim
        axis2 = axis2 % self.ndim
        opp = {axis1: axis2, axis2: axis1}
        axes = [opp.get(i, i) for i in range(self.ndim)]
        return self.relax(self._array.swapaxes(axis1, axis2), self.ann[axes])

    def __matmul__(self, other):
        if isinstance(other, _op.Operator):
            return NotImplemented
        return super().__matmul__(other)


@implements(np.linalg.solve)
def solve(a, b):
    """Solve linear system.

    See also :py:func:`numpy.linalg.solve`.
    """
    if issubclass(type(a), type(b)) or (
        not issubclass(type(b), type(a)) and isinstance(a, AnnotatedArray)
    ):
        restype = type(a)
    else:
        restype = type(b)
    res = AnnotatedArray(np.linalg.solve(np.asanyarray(a), np.asanyarray(b)))
    a_ann = list(getattr(a, "ann", [{}, {}]))
    b_ann = list(getattr(b, "ann", [{}, {}]))
    if np.ndim(b) == np.ndim(a) - 1:
        map(lambda x: x[0].match(x[1]), zip(a_ann[-2::-1], b_ann[-1::-1]))
        del a_ann[-2]
        del b_ann[-1]
    else:
        map(lambda x: x[0].match(x[1]), zip(a_ann[-2::-1], b_ann[-2::-1]))
        del a_ann[-2]
        del b_ann[-2]
        a_ann += [{}]
    res.ann.update(a_ann)
    res.ann.update(b_ann)
    return restype.relax(res)


@implements(np.linalg.lstsq)
def lstsq(a, b, rcond="warn"):
    """Solve linear system using least squares.

    See also :py:func:`numpy.linalg.lstsq`.
    """
    if issubclass(type(a), type(b)) or (
        not issubclass(type(b), type(a)) and isinstance(a, AnnotatedArray)
    ):
        restype = type(a)
    else:
        restype = type(b)
    res = list(np.linalg.lstsq(np.asanyarray(a), np.asanyarray(b), rcond))
    res[0] = AnnotatedArray(res[0])
    a_ann = getattr(a, "ann", ({},))
    b_ann = getattr(b, "ann", ({},))
    a_ann[0].match(b_ann[0])
    res[0].ann[0].update(a_ann[-1])
    if np.ndim(b) == 2:
        res[0].ann[1].update(b_ann[-1])
    res[0] = restype.relax(res[0])
    return tuple(res)


@implements(np.linalg.svd)
def svd(a, full_matrices=True, compute_uv=True, hermitian=False):
    """Compute the singular value decomposition.

    See also :py:func:`numpy.linalg.svd`.
    """
    res = list(np.linalg.svd(np.asanyarray(a), full_matrices, compute_uv, hermitian))
    ann = getattr(a, "ann", ({},))
    if compute_uv:
        res[0] = a.relax(res[0], ann[:-1] + ({},))
        res[1] = a.relax(res[1], ann[:-2] + ({},))
        res[2] = a.relax(res[2], ann[:-2] + ({}, ann[-1]))
        return res
    return a.relax(res, tuple(ann[:-2]) + ({},))


@implements(np.diag)
def diag(a, k=0):
    """Extract diagonal from an array or create an diagonal array.

    See also :py:func:`numpy.diag`.
    """
    res = np.diag(np.asanyarray(a), k)
    ann = a.ann
    if a.ndim == 1:
        ann = (ann[0], copy.copy(ann[0]))
    elif k == 0:
        ann[0].match(ann[1])
        ann = ({**ann[0], **ann[1]},)
    else:
        ann = ()
    return a.relax(res, ann)


@implements(np.tril)
def tril(a, k=0):
    """Get lower trinagular matrix.

    See also :py:func:`numpy.tril`.
    """
    res = np.tril(np.asanyarray(a), k)
    return a.relax(res, getattr(a, "ann", ({},)))


@implements(np.triu)
def triu(a, k=0):
    """Get upper trinagular matrix.

    See also :py:func:`numpy.triu`.
    """
    res = np.triu(np.asanyarray(a), k)
    return a.relax(res, getattr(a, "ann", ({},)))


@implements(np.zeros_like)
def zeros_like(a, dtype=None, order="K", shape=None):
    """Create a numpy array like the given array containing zeros."""
    return np.zeros_like(np.asarray(a), dtype=dtype, order=order, shape=shape)


@implements(np.ones_like)
def ones_like(a, dtype=None, order="K", shape=None):
    """Create a numpy array like the given array containing ones."""
    return np.ones_like(np.asarray(a), dtype=dtype, order=order, shape=shape)



import matplotlib.pyplot as plt
import numpy as np

import treams

k0 = 2 * np.pi / 700
materials = [treams.Material(-16.5 + 1j), treams.Material()]
lmax = mmax = 3
radii = [65, 55]
positions = [[-40, -50, 0], [40, 50, 0]]
lattice_z = treams.Lattice(300)
lattice_x = treams.Lattice(300, "x")
kz = 0.005

sphere = treams.TMatrix.sphere(lmax, k0, radii[0], materials)
chain = sphere.latticeinteraction.solve(lattice_z, kz)
bmax = 1.1 * lattice_z.reciprocal
cwb = treams.CylindricalWaveBasis.diffr_orders(kz, mmax, lattice_z, bmax)
kzs = np.unique(cwb.kz)
chain_tmc = treams.TMatrixC.from_array(chain, cwb)
cylinder = treams.TMatrixC.cylinder(kzs, mmax, k0, radii[1], materials)

cluster = treams.TMatrixC.cluster(
    [chain_tmc, cylinder], positions
).latticeinteraction.solve(lattice_x, 0)
inc = treams.plane_wave(
    [0, np.sqrt(k0 ** 2 - kz ** 2), kz],
    [np.sqrt(0.5), -np.sqrt(0.5)],
    k0=chain.k0,
    material=chain.material,
)
sca = cluster @ inc.expand(cluster.basis)

grid = np.mgrid[0:1, -150:150:16j, -150:150:16j].squeeze().transpose((1, 2, 0))
ex = np.zeros_like(grid[..., 0])
valid = cluster.valid_points(grid, radii)
vals = []
for i, r in enumerate(grid[valid]):
    cwb = treams.CylindricalWaveBasis.default(kzs, 1, positions=[r])
    field = sca.expandlattice(basis=cwb).efield(r)
    vals.append(np.real(inc.efield(r)[0] + field[0]))
ex[valid] = vals
ex[~valid] = np.nan


fig, ax = plt.subplots(figsize=(10, 20))
pcm = ax.pcolormesh(
    grid[0, :, 2], grid[:, 0, 1], ex.T, shading="nearest", vmin=-1, vmax=1,
)
cb = plt.colorbar(pcm)
cb.set_label("$E_x$")
ax.set_xlabel("y (nm)")
ax.set_ylabel("z (nm)")
ax.set_aspect("equal")
fig.show()
